#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <math.h>

// This kernel performs a fused linear transformation and pooling operation while minimizing unnecessary synchronizations.
// It uses __syncthreads() only after critical shared memory operations:
// 1. After loading the input sample into shared memory.
// 2. After computing the full linear transformation into shared memory.
// 3. After writing warp-level partial sums, before the final reduction.
// The pooling reduction is then accomplished using warp shuffle intrinsics to avoid extra barriers.

__global__ void hybrid_min_sync_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,       // batch size
    int inF,     // number of input features
    int outF,    // number of output features
    int kernel_size,
    float scale_factor
) {
    int b = blockIdx.x;
    if (b >= B) return;

    // Shared memory layout:
    // [0, inF)                -> shared copy of input x
    // [inF, inF + outF)       -> results of linear transformation (linear_out)
    // [inF + outF, inF + outF + numWarps) -> warp-level partial sums
    extern __shared__ float sdata[];
    float* shared_x = sdata;             // size: inF
    float* linear_out = shared_x + inF;    // size: outF
    int numWarps = blockDim.x / 32;
    float* warp_sums = linear_out + outF;  // size: numWarps

    // Phase 1: Load input sample x into shared memory
    for (int i = threadIdx.x; i < inF; i += blockDim.x) {
        shared_x[i] = x[b * inF + i];
    }
    __syncthreads(); // Ensure all of shared_x is loaded

    // Phase 2: Compute the linear transformation: linear_out[j] = bias[j] + dot(shared_x, weight_j)
    for (int j = threadIdx.x; j < outF; j += blockDim.x) {
        float acc = bias[j];
        for (int i = 0; i < inF; i++) {
            acc = __fmaf_rn(shared_x[i], weight[j * inF + i], acc);
        }
        linear_out[j] = acc;
    }
    __syncthreads(); // Ensure linear_out is fully computed

    // Phase 3: Perform max pooling over segments of size 'kernel_size'
    int pooled_len = (outF >= kernel_size) ? (1 + (outF - kernel_size) / kernel_size) : 0;
    float local_sum = 0.0f;
    for (int seg = threadIdx.x; seg < pooled_len; seg += blockDim.x) {
        int start = seg * kernel_size;
        float max_val = linear_out[start];
        for (int k = 1; k < kernel_size; k++) {
            max_val = fmaxf(max_val, linear_out[start + k]);
        }
        local_sum += max_val;
    }

    // Phase 4: Intra-warp reduction using warp shuffle to reduce local_sum
    unsigned int mask = 0xffffffff;
    float sum_reg = local_sum;
    for (int offset = 16; offset > 0; offset /= 2) {
        sum_reg += __shfl_down_sync(mask, sum_reg, offset);
    }

    // Phase 5: Each warp leader writes its partial sum to shared memory
    if ((threadIdx.x & 31) == 0) {
        warp_sums[threadIdx.x >> 5] = sum_reg;
    }
    __syncthreads(); // Ensure all warp sums are written

    // Phase 6: Final reduction from warp sums performed by the first warp
    float final_sum = 0.0f;
    if (threadIdx.x < numWarps) {
        final_sum = warp_sums[threadIdx.x];
    }
    if (threadIdx.x < 32) { // First warp reduces the warp_sums
        for (int offset = 16; offset > 0; offset /= 2) {
            final_sum += __shfl_down_sync(mask, final_sum, offset);
        }
    }
    if (threadIdx.x == 0) {
        output[b] = final_sum * scale_factor;
    }
}

// Forward function callable from Python
at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (B, inF)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (outF, inF)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (outF)");

    int B = x.size(0);
    int inF = x.size(1);
    int outF = weight.size(0);

    auto out = torch::empty({B}, x.options());

    // Configure kernel launch
    int threads = 256;
    int numWarps = threads / 32;
    // Shared memory: space for shared_x (inF), linear_out (outF), and warp sums (numWarps)
    size_t sharedMemSize = (inF + outF + numWarps) * sizeof(float);

    // Launch one block per sample
    hybrid_min_sync_kernel<<<B, threads, sharedMemSize>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B, inF, outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Hybrid CUDA forward with minimal synchronizations");
}
